/*-
 * Copyright (C)2004 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2004 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: eol.c,v 1.3 2007/06/06 01:08:33 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "mife.h"
#include "mife_int.h"

typedef struct {
    ssize_t eolp;
    int eolst, eomax;
    const char *eoln;
    const char * const * buf;
    ssize_t eotrix[1];
} eoldetector;
/******************************************/
/* eomax == -1 eoln    */
/* eomax >= 0    eoln       */
/* eolst == -1 eoln       */
/* eolst >= 0           */
/******************************************/
ssize_t
mife_eolinit(mife_descriptor *mife, const char *eoln) {
    int i;

    ifBLIN_QV3(mife->flag) fprintf(stderr, "+mife_eolinit%s%s~\n", eoln ? "=" : "", eoln ? eoln : "");
    if  (eoln) {
        mife->eomax = strlen(eoln);
        if  (mife->eotrilen < mife->eomax) {
            free(mife->eotrix);
            mife->eotrilen = 0;
        }
        if  (!mife->eotrilen && mife->eomax) {
            mife->eotrilen = ((strlen(eoln) * sizeof(ssize_t)) + MINEOLN) & ~MINEOLN;
            mife->eotrix = malloc(mife->eotrilen);
            if  (!mife->eotrix) {
                errno = ENOMEM;
                ifBLIN_QV5(mife->flag) warnx("mife_eolinit");
                ifBLIN_QV3(mife->flag) fprintf(stderr, "-mife_eolinit -1\n");
                return(-1);
        }   }
        mife->eolst = -1;
        mife->eolp = 0;
        mife->eoln = eoln;
        if  (mife->eomax) (mife->eotrix)[0] = -1;
        for (i = 1; i < mife->eomax; i++) {
            if  ((mife->eotrix)[i - 1] < 0) {
                if  (eoln[i] == eoln[0]) (mife->eotrix)[i] = 0;
            } else {
                if  (eoln[i - 1] == eoln[(mife->eotrix)[i - 1]])
                    (mife->eotrix)[i] = (mife->eotrix)[i - 1] + 1;
        }   }
        for (i = 0; i < mife->eomax; i++) {
            if  ((mife->eotrix)[i] < 0) (mife->eotrix)[i] = 0;
            if  (((mife->eotrix)[i] > 0) && (eoln[i] == eoln[(mife->eotrix)[i]])) (mife->eotrix)[i] = 0;
            ifBLIN_QV4(mife->flag) fprintf( stderr, " [%d]%d", i, (mife->eotrix)[i]);
        }
    } else {
        mife->eoln = NULL;
        i = 0;
    }
    ifBLIN_QV3(mife->flag) fprintf( stderr, "\n-mife_eolinit %d\n", i);
    return(i);
}

ssize_t
mife_eoldetect(mife_descriptor *mife) {
    ssize_t r;

    ifBLIN_QV3(mife->flag) fprintf( stderr, "+mife_eoldetect %d[%d]\n", mife->actlen, mife->eolp);
    r = -1;
    if  (mife->eoln && mife->eolst >= 0) {
        while (mife->eolp < mife->actlen && r < 0) {
            ifBLIN_QV4(mife->flag)
                fprintf(stderr, "eolp=%d eolst=%d eomax=%d\n", mife->eolp, mife->eolst, mife->eomax);
            if  (mife->eolst >= mife->eomax) {
                r = mife->eolp;
                mife->eolst = 0;
            } else {
                while (1) {
                    if  (CHARBUF(mife->buffer)[mife->eolp] == (mife->eoln)[mife->eolst]) {
                        ifBLIN_QV5(mife->flag)
                            fprintf(stderr, " + eolp=%d eolst=%d\n", mife->eolp, mife->eolst);
                        mife->eolst++;
                        break;
                    } else if (mife->eolst) {
                        ifBLIN_QV5(mife->flag)
                            fprintf(stderr, " > eolp=%d eolst=%d\n", mife->eolp, mife->eolst);
                        if  (mife->eolst <= (mife->eotrix)[mife->eolst])
                            ifBLIN_QV1(mife->flag)
                                errx( EX_SOFTWARE
                                    , "FATAL: eotrix not decreased %d -> %d"
                                    , mife->eolst, (mife->eotrix)[mife->eolst]
                                    );
                        mife->eolst = (mife->eotrix)[mife->eolst];
                    } else {
                        ifBLIN_QV5(mife->flag)
                            fprintf(stderr, " V eolp=%d eolst=%d\n", mife->eolp, mife->eolst);
                        mife->eolst = 0;
                        break;
                }   }
                mife->eolp++;
    }   }   }
    ifBLIN_QV3(mife->flag) fprintf( stderr, "-mife_eoldetect %d\n", r);
    return(r);
}
